import datetime
from decimal import Decimal
from decimal import ROUND_HALF_UP

import pytest

from tradingbutler.summary import get_position_summaries
from tradingbutler.summary import get_positions

legs = [
    {
        'symbol': 'AAPL',
        'instruction': 'SELL_SHORT',
        'quantity': 16,
        'price': Decimal('29.34'),
        'time': datetime.datetime(2022, 9, 23),
        'order_id': 1
    },
    {
        'symbol': 'AAPL',
        'instruction': 'BUY_TO_COVER',
        'quantity': 16,
        'price': Decimal('23.7'),
        'time': datetime.datetime(2022, 10, 10),
        'order_id': 2
    },
    {
        'symbol': 'ATT',
        'instruction': 'BUY',
        'quantity': 6,
        'price': Decimal('81.55'),
        'time': datetime.datetime(2022, 9, 10),
        'order_id': 3
    },
    {
        "symbol": "ATT",
        'instruction': 'SELL',
        'quantity': 6,
        'price': Decimal('111.25'),
        'time': datetime.datetime(2022, 9, 20),
        'order_id': 4
    },
    {
        'symbol': 'AMZN',
        'instruction': 'BUY',
        'quantity': 12,
        'price': Decimal('75.25'),
        'time': datetime.datetime(2022, 10, 20),
        'order_id': 5
    }
]


def test_get_positions():
    positions = get_positions(legs=legs)
    assert len(positions) == 3
    assert positions[0]['opening']['symbol'] == 'AAPL'
    assert len(positions[0]) == 3
    assert positions[0]['opening']['instruction'] == 'SELL_SHORT'
    assert positions[0]['opening']['price'] == Decimal('29.34')
    assert positions[0]['closing']['instruction'] == 'BUY_TO_COVER'
    assert len(set(positions[0]['order_ids'])) == 2
    assert positions[1]['opening']['symbol'] == 'ATT'
    assert len(positions[1]) == 3
    assert positions[1]['opening']['instruction'] == 'BUY'
    assert positions[1]['opening']['price'] == Decimal('81.55')
    assert positions[1]['closing']['instruction'] == 'SELL'
    assert positions[1]['opening']['quantity'] == Decimal('6.0')
    assert positions[2]['opening']['symbol'] == 'AMZN'
    assert len(positions[2]) == 2
    assert positions[2]['opening']['instruction'] == 'BUY'
    assert positions[2]['opening']['price'] == Decimal('75.25')
    assert len(set(positions[2]['order_ids'])) == 1


def test_get_position_summaries():
    positions = get_positions(legs)
    position_summaries = get_position_summaries(positions)
    assert len(position_summaries) == 3
    assert position_summaries[0]['symbol'] == 'AAPL'
    assert position_summaries[1]['symbol'] == 'ATT'
    assert position_summaries[1]['average_price'] == Decimal('81.55')
    assert position_summaries[1]['risk'] == Decimal('489.300')
    assert position_summaries[1]['days'] == 10
    assert position_summaries[1]['profit'] == Decimal('178.2000')
    assert position_summaries[1]['profit_percentage'] == Decimal('36.42')
    assert position_summaries[1]['entry_date'] == datetime.date(2022, 9, 10)
    assert position_summaries[1]['exit_date'] == datetime.date(2022, 9, 20)
    assert position_summaries[1]['exit_price'] == Decimal('111.25')
    assert position_summaries[1]['direction'] == 'Long'
    assert position_summaries[2]['symbol'] == 'AMZN'
    assert position_summaries[2]['exit_price'] == None
    assert position_summaries[2]['exit_date'] == None
    assert position_summaries[2]['profit'] == None
    assert position_summaries[2]['profit_percentage'] == None


def test_partial_closing():
    data = [
        {
            'symbol': 'META',
            'instruction': 'BUY',
            'quantity': 10,
            'price': Decimal('241.34'),
            'time': datetime.datetime(2022, 10, 13),
            'order_id': 1
        },
        {
            'symbol': 'META',
            'instruction': 'SELL',
            'quantity': 5,
            'price': Decimal('232.11'),
            'time': datetime.datetime(2022, 10, 20),
            'order_id': 2
        },
    ]

    positions = get_positions(legs=data)
    assert len(positions) == 2
    expected = [
        {
            'closing':
                {
                    'instruction': 'SELL',
                    'price': Decimal('232.1100'),
                    'quantity': 5,
                    'symbol': 'META',
                    'time': datetime.datetime(2022, 10, 20)
                },
            'opening':
                {
                    'instruction': 'BUY',
                    'price': Decimal('241.3400'),
                    'quantity': 5,
                    'risk': Decimal('1206.7000'),
                    'symbol': 'META',
                    'time': datetime.datetime(2022, 10, 13)
                },
            'order_ids':
                [1, 1, 1, 1, 1, 2]
        },
        {
            'opening':
                {
                    'instruction': 'BUY',
                    'price': Decimal('241.3400'),
                    'quantity': 5,
                    'risk': Decimal('1206.7000'),
                    'symbol': 'META',
                    'time': datetime.datetime(2022, 10, 13)
                },
            'order_ids':
                [1, 1, 1, 1, 1]
        }
    ]
    assert positions == expected


def test_partial_closing_multiple_buys():
    data = [
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 12,
            'price': Decimal('72.01'),
            'time': datetime.datetime(2022, 9, 20),
            'order_id': 1
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 6,
            'price': Decimal('54.21'),
            'time': datetime.datetime(2022, 10, 4),
            'order_id': 2
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 4,
            'price': Decimal('50.21'),
            'time': datetime.datetime(2022, 10, 7),
            'order_id': 3
        },
        {
            'symbol': 'GE',
            'instruction': 'SELL',
            'quantity': 11,
            'price': Decimal('65.14'),
            'time': datetime.datetime(2022, 10, 17),
            'order_id': 4
        }
    ]
    positions = get_positions(legs=data)
    assert len(positions) == 2
    expected = [
        {
            'closing':
                {
                    'instruction': 'SELL',
                    'price': Decimal('65.14'),
                    'quantity': 11,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 17)
                },
            'opening':
                {
                    'instruction': 'BUY',
                    'price': Decimal('72.0100'),
                    'quantity': 11,
                    'risk': Decimal('792.1100'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4]
        },
        {
            'opening':
                {
                    'instruction': 'BUY',
                    'price': ((1 * Decimal('72.01') + 6 * Decimal('54.21') + 4 * Decimal('50.21')) / 11).quantize(Decimal('.0001'), ROUND_HALF_UP),
                    'quantity': 11,
                    'risk': Decimal('598.1100'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3]
        }
    ]
    assert positions == expected


def test_only_buy():
    data = [
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 12,
            'price': Decimal('72.01'),
            'time': datetime.datetime(2022, 9, 20),
            'order_id': 1
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 6,
            'price': Decimal('54.21'),
            'time': datetime.datetime(2022, 10, 4),
            'order_id': 2
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 4,
            'price': Decimal('50.21'),
            'time': datetime.datetime(2022, 10, 7),
            'order_id': 3
        },
    ]
    positions = get_positions(legs=data)
    assert len(positions) == 1
    expected = [
        {
            'opening':
                {
                    'instruction': 'BUY',
                    'price': ((12 * Decimal('72.01') + 6 * Decimal('54.21') + 4 * Decimal('50.21')) / 22).quantize(Decimal('.0001'), ROUND_HALF_UP),
                    'quantity': 22,
                    'risk': Decimal('1390.2200'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3]
        }
    ]
    assert positions == expected


def test_partial_closing_multiple_sells():
    data = [
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 12,
            'price': Decimal('72.01'),
            'time': datetime.datetime(2022, 9, 20),
            'order_id': 1
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 6,
            'price': Decimal('54.21'),
            'time': datetime.datetime(2022, 10, 4),
            'order_id': 2
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 4,
            'price': Decimal('50.21'),
            'time': datetime.datetime(2022, 10, 7),
            'order_id': 3
        },
        {
            'symbol': 'GE',
            'instruction': 'SELL',
            'quantity': 11,
            'price': Decimal('65.14'),
            'time': datetime.datetime(2022, 10, 17),
            'order_id': 4
        },
        {
            'symbol': 'GE',
            'instruction': 'SELL',
            'quantity': 5,
            'price': Decimal('69.33'),
            'time': datetime.datetime(2022, 10, 21),
            'order_id': 5
        },
        {
            'symbol': 'GE',
            'instruction': 'SELL',
            'quantity': 4,
            'price': Decimal('70.55'),
            'time': datetime.datetime(2022, 10, 26),
            'order_id': 6
        }
    ]

    positions = get_positions(legs=data)
    assert len(positions) == 4
    expected = [
        {
            'closing':
                {
                    'instruction': 'SELL',
                    'price': Decimal('65.14'),
                    'quantity': 11,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 17)
                },
            'opening':
                {
                    'instruction': 'BUY',
                    'price': Decimal('72.0100'),
                    'quantity': 11,
                    'risk': Decimal('792.1100'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4]
        },
        {
            'closing':
                {
                    'instruction': 'SELL',
                    'price': Decimal('69.33'),
                    'quantity': 5,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 21)
                },
            'opening':
                {
                    'instruction': 'BUY',
                    'price': ((1 * Decimal('72.01') + 4 * Decimal('54.21')) / 5).quantize(Decimal('.0001'), ROUND_HALF_UP),
                    'quantity': 5,
                    'risk': Decimal('288.8500'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 2, 2, 2, 2, 5]
        },
        {
            'closing':
                {
                    'instruction': 'SELL',
                    'price': Decimal('70.55'),
                    'quantity': 4,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 26)
                },
            'opening':
                {
                    'instruction': 'BUY',
                    'price': ((2 * Decimal('54.21') + 2 * Decimal('50.21')) / 4).quantize(Decimal('.0001'), ROUND_HALF_UP),
                    'quantity': 4,
                    'risk': Decimal('208.8400'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 4)
                },
            'order_ids':
                [2, 2, 3, 3, 6]
        },
        {
            'opening':
                {
                    'instruction': 'BUY',
                    'price': Decimal('50.2100'),
                    'quantity': 2,
                    'risk': Decimal('100.4200'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 7)
                },
            'order_ids':
                [3, 3]
        }
    ]
    assert positions == expected


def test_sell_short_fully_closed():
    data = [
        {
            'symbol': 'GE',
            'instruction': 'SELL_SHORT',
            'quantity': 12,
            'price': Decimal('72.01'),
            'time': datetime.datetime(2022, 9, 20),
            'order_id': 1
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY_TO_COVER',
            'quantity': 12,
            'price': Decimal('54.21'),
            'time': datetime.datetime(2022, 10, 4),
            'order_id': 2
        },
    ]
    positions = get_positions(legs=data)
    assert len(positions) == 1
    expected = [
        {
            'closing':
                {
                    'instruction': 'BUY_TO_COVER',
                    'price': Decimal('54.21'),
                    'quantity': 12,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 4)
                },
            'opening':
                {
                    'instruction': 'SELL_SHORT',
                    'price': Decimal('72.0100'),
                    'quantity': 12,
                    'risk': Decimal('864.1200'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2]
        },
    ]
    assert positions == expected


def test_partial_closing_sell_short_multiple_positions():
    data = [
        {
            'symbol': 'GE',
            'instruction': 'SELL_SHORT',
            'quantity': 7,
            'price': Decimal('72.01'),
            'time': datetime.datetime(2022, 9, 20),
            'order_id': 1
        },
        {
            'symbol': 'GE',
            'instruction': 'SELL_SHORT',
            'quantity': 8,
            'price': Decimal('77.22'),
            'time': datetime.datetime(2022, 9, 28),
            'order_id': 2
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY_TO_COVER',
            'quantity': 10,
            'price': Decimal('54.21'),
            'time': datetime.datetime(2022, 10, 4),
            'order_id': 3
        },
        {
            'symbol': 'GE',
            'instruction': 'BUY_TO_COVER',
            'quantity': 3,
            'price': Decimal('60.33'),
            'time': datetime.datetime(2022, 10, 8),
            'order_id': 4
        },
    ]
    positions = get_positions(legs=data)
    assert len(positions) == 3
    expected = [
        {
            'closing':
                {
                    'instruction': 'BUY_TO_COVER',
                    'price': Decimal('54.21'),
                    'quantity': 10,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 4)
                },
            'opening':
                {
                    'instruction': 'SELL_SHORT',
                    'price': ((7 * Decimal('72.01') + 3 * Decimal('77.22')) / 10).quantize(Decimal('.0001'), ROUND_HALF_UP),
                    'quantity': 10,
                    'risk': Decimal('735.7300'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 20)
                },
            'order_ids':
                [1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 3]
        },
        {
            'closing':
                {
                    'instruction': 'BUY_TO_COVER',
                    'price': Decimal('60.33'),
                    'quantity': 3,
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 10, 8)
                },
            'opening':
                {
                    'instruction': 'SELL_SHORT',
                    'price': Decimal('77.2200'),
                    'quantity': 3,
                    'risk': Decimal('231.6600'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 28)
                },
            'order_ids':
                [2, 2, 2, 4]
        },
        {
            'opening':
                {
                    'instruction': 'SELL_SHORT',
                    'price': Decimal('77.2200'),
                    'quantity': 2,
                    'risk': Decimal('154.4400'),
                    'symbol': 'GE',
                    'time': datetime.datetime(2022, 9, 28)
                },
            'order_ids':
                [2, 2]
        }
    ]
    assert positions == expected


def test_closing_more_shares_than_open_fails():
    data = [
        {
            'symbol': 'GE',
            'instruction': 'BUY',
            'quantity': 5,
            'price': Decimal('72.01'),
            'time': datetime.datetime(2022, 9, 20),
            'order_id': 1
        },
        {
            'symbol': 'GE',
            'instruction': 'SELL',
            'quantity': 10,
            'price': Decimal('54.21'),
            'time': datetime.datetime(2022, 10, 4),
            'order_id': 2
        },
    ]
    with pytest.raises(ValueError) as excinfo:
        get_positions(legs=data)
    assert excinfo.value.args[0] == 'closing more shares than open'



import datetime
import json
from collections import deque
from decimal import ROUND_HALF_UP
from decimal import Decimal

from dateutil.parser import parse
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

env = Environment(
    loader=PackageLoader(__package__),
    autoescape=select_autoescape()
)


class BaseTradeImporter:
    def __init__(self, json_string, descending=False):
        parsed = json.loads(json_string, parse_float=Decimal)
        if isinstance(parsed, list):
            self.trades = parsed
        elif isinstance(parsed, dict):
            self.trades = [parsed]
        else:
            raise ValueError
        self.legs = self.get_legs(self.trades)
        self.descending = descending

    @classmethod
    def from_path(cls, path):
        json_string = cls.read_file(cls, path)
        return cls(json_string)

    def read_file(self, path):
        with open(path, 'r') as f:
            json_string = f.read()
        return json_string

    def get_legs(self, trades):
        raise NotImplementedError


class TdaTradeImporter(BaseTradeImporter):
    def get_legs(self, trades):
        if not self.descending:
            trades = reversed(trades)
        legs = []
        for trade in trades:
            instruction = None
            symbol = None
            order_id = trade['orderId']
            for order_leg in trade['orderLegCollection']:
                instruction = order_leg['instruction']
                symbol = order_leg['instrument']['symbol']

            for order_activity in trade['orderActivityCollection']:
                for execution_leg in order_activity['executionLegs']:
                    legs.append({
                        'quantity': int(execution_leg['quantity']),
                        'price': execution_leg['price'],
                        'time': parse(execution_leg['time']),
                        'instruction': instruction,
                        'symbol': symbol,
                        'order_id': order_id
                    })

        return legs


def get_positions(legs):
    positions = []
    current_positions = {}
    for leg in legs:
        if not leg['symbol'] in current_positions:
            current_positions[leg['symbol']] = deque()
        if leg['instruction'] in ['BUY', 'SELL_SHORT']:
            quantity = leg['quantity']
            for i in range(quantity):
                symbol = leg['symbol']
                instruction = leg['instruction']
                price = leg['price']
                time = leg['time']
                order_id = leg['order_id']
                share = {
                    'symbol': symbol,
                    'instruction': instruction,
                    'price': price,
                    'time': time,
                    'order_id': order_id
                }
                current_positions[leg['symbol']].append(share)
        elif leg['instruction'] in ['SELL', 'BUY_TO_COVER']:
            closing_quantity = leg['quantity']
            closing_instruction = leg['instruction']
            closing_price = leg['price']
            closing_time = leg['time']
            to_be_closed = []
            for i in range(closing_quantity):
                try:
                    share = current_positions[leg['symbol']].popleft()
                except IndexError:
                    raise ValueError('closing more shares than open')
                to_be_closed.append(share)
            risk = sum(share['price'] for share in to_be_closed)
            average_price = (risk / closing_quantity).quantize(Decimal('.0001'), ROUND_HALF_UP)
            symbol = to_be_closed[0]['symbol']
            instruction = to_be_closed[0]['instruction']
            time = to_be_closed[0]['time']
            opening_leg = {
                'symbol': symbol,
                'instruction': instruction,
                'quantity': len(to_be_closed),
                'price': average_price,
                'time': time,
                'risk': risk.quantize(Decimal('.0001'), ROUND_HALF_UP)
            }
            closing_leg = {
                'symbol': symbol,
                'instruction': closing_instruction,
                'quantity': closing_quantity,
                'price': closing_price,
                'time': closing_time
            }
            order_ids = [share['order_id'] for share in to_be_closed]
            order_ids.append(leg['order_id'])
            order_ids = sorted(set(order_ids))
            position = {
                'opening': opening_leg,
                'closing': closing_leg,
                'order_ids': order_ids
            }
            positions.append(position)

    # Add the open positions
    for symbol, q in current_positions.items():
        if len(q):
            still_open = []
            for i in range(len(q)):
                still_open.append(q.popleft())
            risk = sum(share['price'] for share in still_open)
            average_price = (risk / len(still_open)).quantize(Decimal('.0001'), ROUND_HALF_UP)
            symbol = still_open[0]['symbol']
            instruction = still_open[0]['instruction']
            time = still_open[0]['time']
            opening_leg = {
                'symbol': symbol,
                'instruction': instruction,
                'quantity': len(still_open),
                'price': average_price,
                'time': time,
                'risk': risk.quantize(Decimal('.0001'), ROUND_HALF_UP)
            }
            order_ids = [share['order_id'] for share in still_open]
            order_ids = sorted(set(order_ids))
            position = {
                'opening': opening_leg,
                'order_ids': order_ids
            }
            positions.append(position)
    return positions


def get_position_summaries(positions, current_date=None):
    if current_date is None:
        current_date = datetime.datetime.now().date()
    position_summaries = []
    for position in positions:
        position_summary = {}
        position_summaries.append(position_summary)
        order_ids = position['order_ids']
        number_legs = len(set(order_ids))
        risk = position['opening']['risk']
        quantity = position['opening']['quantity']
        price = position['opening']['price']
        symbol = position['opening']['symbol']
        direction = 'Long' if position['opening']['instruction'] == 'BUY' else 'Short'
        entry_date = position['opening']['time'].date()
        if 'closing' in position:
            size = quantity * position['closing']['price']
            exit_price = size / quantity
            rounded_exit_price = exit_price.quantize(Decimal('0.0001'), rounding=ROUND_HALF_UP)
            profit_percentage = ((exit_price / price) - 1) * 100
            profit_percentage = profit_percentage.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP)
            profit = size - risk
            rounded_profit = profit.quantize(Decimal('0.0001'), rounding=ROUND_HALF_UP)
            exit_date = position['closing']['time'].date()
            days = (exit_date - entry_date).days
            if 'BUY_TO_COVER' in position['closing']['instruction']:
                rounded_profit = rounded_profit * -1
                profit_percentage = profit_percentage * -1
        else:
            rounded_exit_price = None
            exit_date = None
            rounded_profit = None
            profit_percentage = None
            days = (current_date - entry_date).days
        position_summary.update({
            'symbol': symbol,
            'risk': risk,
            'entry_date': entry_date,
            'average_price': price,
            'exit_price': rounded_exit_price,
            'exit_date': exit_date,
            'days': days,
            'quantity': quantity,
            'direction': direction,
            'profit': rounded_profit,
            'profit_percentage': profit_percentage,
            'number_legs': number_legs,
            'order_ids': order_ids
        })
    return position_summaries


def write_output(position_summaries, keys=None, template_file='template.html', output_file='output.html'):
    if keys is None:
        keys = ['symbol', 'direction', 'entry_date', 'average_price', 'exit_price', 'exit_date', 'days', 'quantity',
                'risk', 'profit_percentage', 'profit', 'number_legs']
    template = env.get_template(template_file)
    output = template.render(position_summaries=position_summaries, keys=keys)
    with open(output_file, 'w') as f:
        f.write(output)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tradingbutler', 'tradingbutler.tests']

package_data = \
{'': ['*'], 'tradingbutler': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0', 'python-dateutil>=2.8.2,<3.0.0']

setup_kwargs = {
    'name': 'tradingbutler',
    'version': '0.1.10',
    'description': '',
    'long_description': '# tradingbutler\n\norganizes data from trades and turn into statistics',
    'author': 'Robin Huo',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Robinhuo1/tradingbutler',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
